/* David Leonard, 2002. Public domain. */
/* $Id: proto_dccm.c,v 1.1.1.1 2002/01/29 04:48:28 d Exp $ */

#include <stdio.h>
#include <sys/types.h>
#include <machine/endian.h>

/* Protocol decoder for Direct Connection Client Manager (Activesync) */

#include "proto.h"

static void
utoc(s, len, us)
	char *s;
	size_t len;
	const u_int16_t *us;
{
	const u_int16_t *up;
	char *p;

	for (p = s, up = us; len-- && *up; p++, up++)
		*p = letoh16(*up) & 0xff;
	if (len)
		*p++ = '\0';
}


static void
decode(state, context, reverse, data, datalen)
	void *state;
	void *context;
	int reverse;
	const unsigned char *data;
	int datalen;
{
	u_int32_t u32, len, sz;
	int off1, off2, off3;
	char s[256];
	const unsigned char *d, *d1, *dend;

	dend = data + datalen;
	d = data;

	if (d + sizeof u32 > dend)
		return;
	memcpy(&u32, d, sizeof u32);
	sz = letoh32(u32);
	d += sizeof u32;

	if (sz == 0) {
		proto_print(context, d, "<dccm nop>");
		return;
	}
	if (sz == 0x12345678) {
		proto_print(context, d, "<dccm ping>");
		return;
	}
	if (sz > 0x10000) {
		proto_print(context, d, "<dccm challenge 0x%02x>",
		    (int)(sz & 0xff));
		return;
	}

	if (d + sz < dend)
		return;

	d1 = d;
	memcpy(&u32, d, sizeof u32);
	len = letoh32(u32);			/* subheader len? */
	d += sizeof u32;

	if (len != 0x24)
		proto_print(context, dend, "<dccm unknown>");

	memcpy(&u32, d1 + 0x18, sizeof u32); off1 = letoh32(u32);
	memcpy(&u32, d1 + 0x1c, sizeof u32); off2 = letoh32(u32);
	memcpy(&u32, d1 + 0x20, sizeof u32); off3 = letoh32(u32);

	if (off1 != len || off2 <= off1 || off3 <= off2 || off3 >= (dend - d1))
		proto_print(context, dend, "<dccm unknown>");

	proto_print(context, d, "<dccm info %x %x %x>", off1, off2, off3);

	d = d1 + off1;
	proto_print(context, d, "");

	utoc(s, sizeof s, d1 + off1);
	d = d1 + off2;
	proto_print(context, d, "name: \"%s\"", s);

	utoc(s, sizeof s, d1 + off2);
	d = d1 + off3;
	proto_print(context, d, "class: \"%s\"", s);

	utoc(s, sizeof s, d1 + off3);
	proto_print(context, dend, "hw: \"%s\"", s);
}

struct proto proto_dccm = {
	5679, 	/* port */
	0,	/* state */
	0,	/* init */
	decode,	/* decode */
	0	/* final */
};
