/* David Leonard, 2002. Public domain. */
/* $Id: proto_x11.c,v 1.3 2002/07/30 07:32:47 d Exp $ */

#include <stdio.h>
#include <sys/types.h>
#include <machine/endian.h>

#include <X11/X.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>

/* Protocol decoder for X11 */

#include "proto.h"

struct x11state {
	enum { NOTCONN, CONNECTING, CONNECTING2, CONNECTED, FAILED } conn;
	u_int8_t	byteorder;
	unsigned long	requestlen;
	unsigned long	sconnlen;
	u_int16_t	reqseq;
};

#define U16(v)	(xs->byteorder == 'B' ? betoh16(v) : letoh16(v))
#define U32(v)	(xs->byteorder == 'B' ? betoh32(v) : letoh32(v))

#define Authenticate 2		/* not actually in any header! */

static const char *reqname[] = {
        NULL, "X_CreateWindow", "X_ChangeWindowAttributes",
        "X_GetWindowAttributes", "X_DestroyWindow", "X_DestroySubwindows",
        "X_ChangeSaveSet", "X_ReparentWindow", "X_MapWindow",
        "X_MapSubwindows", "X_UnmapWindow", "X_UnmapSubwindows",
        "X_ConfigureWindow", "X_CirculateWindow", "X_GetGeometry",
        "X_QueryTree", "X_InternAtom", "X_GetAtomName", "X_ChangeProperty",
        "X_DeleteProperty", "X_GetProperty", "X_ListProperties",
        "X_SetSelectionOwner", "X_GetSelectionOwner", "X_ConvertSelection",
        "X_SendEvent", "X_GrabPointer", "X_UngrabPointer",
        "X_GrabButton", "X_UngrabButton", "X_ChangeActivePointerGrab",
        "X_GrabKeyboard", "X_UngrabKeyboard", "X_GrabKey",
        "X_UngrabKey", "X_AllowEvents", "X_GrabServer", "X_UngrabServer",
        "X_QueryPointer", "X_GetMotionEvents", "X_TranslateCoords",
        "X_WarpPointer", "X_SetInputFocus", "X_GetInputFocus",
        "X_QueryKeymap", "X_OpenFont", "X_CloseFont", "X_QueryFont",
        "X_QueryTextExtents", "X_ListFonts", "X_ListFontsWithInfo",
        "X_SetFontPath", "X_GetFontPath", "X_CreatePixmap",
        "X_FreePixmap", "X_CreateGC", "X_ChangeGC", "X_CopyGC",
        "X_SetDashes", "X_SetClipRectangles", "X_FreeGC", "X_ClearArea",
        "X_CopyArea", "X_CopyPlane", "X_PolyPoint", "X_PolyLine",
        "X_PolySegment", "X_PolyRectangle", "X_PolyArc", "X_FillPoly",
        "X_PolyFillRectangle", "X_PolyFillArc", "X_PutImage",
        "X_GetImage", "X_PolyText", "X_PolyText", "X_ImageText",
        "X_ImageText", "X_CreateColormap", "X_FreeColormap",
        "X_CopyColormapAndFree", "X_InstallColormap",
        "X_UninstallColormap", "X_ListInstalledColormaps",
        "X_AllocColor", "X_AllocNamedColor", "X_AllocColorCells",
        "X_AllocColorPlanes", "X_FreeColors", "X_StoreColors",
        "X_StoreNamedColor", "X_QueryColors", "X_LookupColor",
        "X_CreateCursor", "X_CreateGlyphCursor", "X_FreeCursor",
        "X_RecolorCursor", "X_QueryBestSize", "X_QueryExtension",
        "X_ListExtensions", "X_ChangeKeyboardMapping",
        "X_GetKeyboardMapping", "X_ChangeKeyboardControl",
        "X_GetKeyboardControl", "X_Bell", "X_ChangePointerControl",
        "X_GetPointerControl", "X_SetScreenSaver", "X_GetScreenSaver",
        "X_ChangeHosts", "X_ListHosts", "X_SetAccessControl",
        "X_SetCloseDownMode", "X_KillClient", "X_RotateProperties",
        "X_ForceScreenSaver", "X_SetPointerMapping", "X_GetPointerMapping",
        "X_SetModifierMapping", "X_GetModifierMapping", NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, "X_NoOperation"
};

const char *eventname[] = {
	NULL, NULL,
        "KeyPress", "KeyRelease", "ButtonPress", "ButtonRelease",
        "MotionNotify", "EnterNotify", "LeaveNotify", "FocusIn",
        "FocusOut", "KeymapNotify", "Expose", "GraphicsExpose",
        "NoExpose", "VisibilityNotify", "CreateNotify", "DestroyNotify",
        "UnmapNotify", "MapNotify", "MapRequest", "ReparentNotify",
        "ConfigureNotify", "ConfigureRequest", "GravityNotify",
        "ResizeRequest", "CirculateNotify", "CirculateRequest",
        "PropertyNotify", "SelectionClear", "SelectionRequest",
        "SelectionNotify", "ColormapNotify", "ClientMessage",
        "MappingNotify"
};

const char *errorname[] = {
        NULL, "BadRequest", "BadValue", "BadWindow", "BadPixmap",
        "BadAtom", "BadCursor", "BadFont", "BadMatch", "BadDrawable",
        "BadAccess", "BadAlloc", "BadColor", "BadGC", "BadIDChoice",
        "BadName", "BadLength", "BadImplementation"
};

const char *atomname[] = {
        NULL, "XA_PRIMARY", "XA_SECONDARY", "XA_ARC", "XA_ATOM",
        "XA_BITMAP", "XA_CARDINAL", "XA_COLORMAP", "XA_CURSOR",
        "XA_CUT_BUFFER0", "XA_CUT_BUFFER1", "XA_CUT_BUFFER2",
        "XA_CUT_BUFFER3", "XA_CUT_BUFFER4", "XA_CUT_BUFFER5",
        "XA_CUT_BUFFER6", "XA_CUT_BUFFER7", "XA_DRAWABLE", "XA_FONT",
        "XA_INTEGER", "XA_PIXMAP", "XA_POINT", "XA_RECTANGLE",
        "XA_RESOURCE_MANAGER", "XA_RGB_COLOR_MAP", "XA_RGB_BEST_MAP",
        "XA_RGB_BLUE_MAP", "XA_RGB_DEFAULT_MAP", "XA_RGB_GRAY_MAP",
        "XA_RGB_GREEN_MAP", "XA_RGB_RED_MAP", "XA_STRING",
        "XA_VISUALID", "XA_WINDOW", "XA_WM_COMMAND", "XA_WM_HINTS",
        "XA_WM_CLIENT_MACHINE", "XA_WM_ICON_NAME", "XA_WM_ICON_SIZE",
        "XA_WM_NAME", "XA_WM_NORMAL_HINTS", "XA_WM_SIZE_HINTS",
        "XA_WM_ZOOM_HINTS", "XA_MIN_SPACE", "XA_NORM_SPACE",
        "XA_MAX_SPACE", "XA_END_SPACE", "XA_SUPERSCRIPT_X",
        "XA_SUPERSCRIPT_Y", "XA_SUBSCRIPT_X", "XA_SUBSCRIPT_Y",
        "XA_UNDERLINE_POSITION", "XA_UNDERLINE_THICKNESS",
        "XA_STRIKEOUT_ASCENT", "XA_STRIKEOUT_DESCENT", "XA_ITALIC_ANGLE",
        "XA_X_HEIGHT", "XA_QUAD_WIDTH", "XA_WEIGHT", "XA_POINT_SIZE",
        "XA_RESOLUTION", "XA_COPYRIGHT", "XA_NOTICE", "XA_FONT_NAME",
        "XA_FAMILY_NAME", "XA_FULL_NAME", "XA_CAP_HEIGHT",
        "XA_WM_CLASS", "XA_WM_TRANSIENT_FOR"
};
#define lengthof(a)	(sizeof (a) / sizeof (a)[0])
#define ATOMNAME(a) ((a) < lengthof(atomname) ? atomname[a] : "?")

static void
init(state)
	void *state;
{
	struct x11state *xs = (struct x11state *)state;

	xs->conn = NOTCONN;
	xs->requestlen = 0;
}

static void
decode(state, context, fromclient, data, datalen)
	void *state;
	void *context;
	int fromclient;
	const unsigned char *data;
	int datalen;
{
	struct x11state *xs = (struct x11state *)state;
	const unsigned char *d, *dend;

	dend = data + datalen;
	d = data;

	if (dend == data)
		return;

	/* Dead connections */
	if (xs->conn == FAILED) {
		proto_print(context, dend, "<failed>");
		return;
	}

	/* Initial connection from client */
	if (fromclient && xs->conn == NOTCONN) {
		const xConnClientPrefix *p = (const xConnClientPrefix *)data;
		int aplen, aslen;

		if (dend < data + sz_xConnClientPrefix)
			return;
		xs->byteorder = p->byteOrder;
		xs->reqseq = 0;
		xs->conn = CONNECTING;

		aplen = (U16(p->nbytesAuthProto)+3)&~3;
		aslen = (U16(p->nbytesAuthString)+3)&~3;

		d = data + sz_xConnClientPrefix;

		proto_print(context, d, "<ConnClient> %s X%d.%d",
			xs->byteorder == 'B' ? "big-endian" : "little-endian",
			U16(p->majorVersion), U16(p->minorVersion));
		proto_print(context, d + aplen, " auth proto: %.*s", aplen, d);
		d += aplen;
		proto_print(context, d + aslen, " auth string: %s", 
		    quotestr(d, aslen));
		d += aslen;

		return;
	}

	/* Out-of-order server data? */
	if (!fromclient && xs->conn == NOTCONN) {
		proto_print(context, dend, "<notconn?>");
		return;
	}

	/* Ignore client authentication stuff */
	if (fromclient && xs->conn == CONNECTING) {
		proto_print(context, dend, "");
		return;
	}

	/* Server's initial response */
	if (!fromclient && xs->conn == CONNECTING) {
		const xConnSetupPrefix *p = (const xConnSetupPrefix *)data;
		d = data + sz_xConnSetupPrefix;
		if (dend < d)
			return;
		if (dend < (d + p->lengthReason))
			return;
		proto_print(context, d, "<ConnSetup> %s X%u.%u",
			p->success == xTrue ? "Success" :
			p->success == xFalse ? "Failed" :
			p->success == Authenticate ? "Authenticate" : "?",
			U16(p->majorVersion),
			U16(p->minorVersion));
		if (p->lengthReason)
			proto_print(context, d + p->lengthReason, "reason: %s",
				quotestr(d, p->lengthReason));
		switch (p->success) {
		case xTrue /* Success */:
			xs->sconnlen = U16(p->length) * 4;
			xs->conn = CONNECTING2;
			break;
		case Authenticate:
			break;
		case xFalse /* Failed */:
		default:
			xs->conn = FAILED;
			break;
		}
		return;
	}

	/* Server's response to complete the session establishment */
	if (!fromclient && xs->conn == CONNECTING2) {
		const xConnSetup *p = (const xConnSetup *)data;
		if (dend < data + xs->sconnlen)
			return;
		proto_print(context, (void *)&p->motionBufferSize, 
		    " release %u ridBase 0x%08x ridMask 0x%08x",
		    U32(p->release), U32(p->ridBase), U32(p->ridMask));
		proto_print(context, &p->numRoots, 
		    " motionBufferSize %u maxRequestSize %u",
		    U32(p->motionBufferSize), U16(p->maxRequestSize));
		proto_print(context, (void *)&p->bitmapScanlineUnit,
		    " roots %u formats %u imageBO %s pixelBO %s",
		    p->numRoots, p->numFormats, 
		    p->imageByteOrder == LSBFirst ? "LSBFirst" :
		    p->imageByteOrder == MSBFirst ? "MSBFirst" : "?",
		    p->bitmapBitOrder == LSBFirst ? "LSBFirst" :
		    p->bitmapBitOrder == MSBFirst ? "MSBFirst" : "?");
		proto_print(context, data + sz_xConnSetup,
		    " scanlineUnit %u scanlinePad %u minKey %u maxKey %u",
		    p->bitmapScanlineUnit, p->bitmapScanlinePad,
		    p->minKeyCode, p->maxKeyCode);
		if (xs->sconnlen > sz_xConnSetup)
			proto_print(context, data + xs->sconnlen, "");
		xs->conn = CONNECTED;
		return;
	}

	/* Client sends a request */
	if (fromclient && xs->conn == CONNECTED) {
		const xReq *xr = (const xReq *)data;
		const xChangePropertyReq *cpr =(const xChangePropertyReq *)data;
		int len, ulen;

		if (dend < data + sz_xReq)
			return;
		len = U16(xr->length) * 4;
		if (dend < data + len)
			return;
		xs->reqseq++;

		if (xr->reqType < 128 && reqname[xr->reqType])
			proto_print(context, data + sz_xReq,
			    "#%u <%s>",
			    xs->reqseq, reqname[xr->reqType]);
		else
			proto_print(context, data + sz_xReq,
			    "#%u <unknown request %d>",
			    xs->reqseq,
			    xr->reqType);

		switch (xr->reqType) {
		case X_ChangeProperty:
			d = data + sz_xChangePropertyReq;
			proto_print(context, (void*)&cpr->type,
				" window %08x property %s",
				U32(cpr->window),
				ATOMNAME(U32(cpr->property)));
			ulen = U32(cpr->nUnits);
			proto_print(context, d,
				" type %s format %u length %u",
				ATOMNAME(U32(cpr->type)),
				cpr->format, ulen);
			if (U32(cpr->type) == XA_STRING && cpr->format == 8)
				proto_print(context, d + ulen, 
				    " string: %s", quotestr(d, ulen));
			break;
		}

		proto_print(context, data + len, NULL);

		return;
	}

	/* Server sends a response or an event */
	if (!fromclient && xs->conn == CONNECTED) {
		const xGenericReply *xr = (const xGenericReply *)data;
		int len;

		if (dend < data + 1)
			return;
		if (data[0] == X_Error) {
			const xError *xe = (const xError *)data;
			if (dend < data + sz_xError)
				return;
			if (xe->errorCode < (sizeof errorname 
			    / sizeof errorname[0]) && errorname[xe->errorCode])
			    proto_print(context, (void *)&xe->resourceID,
				"<X_Error %s> seq #%u",
				errorname[xe->errorCode],
				U16(xe->sequenceNumber));
			else
			    proto_print(context, (void *)&xe->resourceID,
				"<X_Error> errorCode %u seq #%u",
				xe->errorCode, U16(xe->sequenceNumber));
			proto_print(context, data + sz_xError,
				"resourceID %08x minorCode %d majorCode %d",
				U32(xe->resourceID),
				U16(xe->minorCode),
				xe->majorCode);
			return;
		}

		if (data[0] == X_Reply) {
			if (dend < data + sz_xGenericReply)
				return;
			len = sz_xGenericReply + U16(xr->length) * 4;
			if (dend < data + len)
				return;

			proto_print(context, data + sz_xGenericReply,
			    "<X_Reply #%u>", U16(xr->sequenceNumber));

			if (len > sz_xGenericReply)
				proto_print(context, data + len, "");
			return;
		}

		if (data[0] < LASTEvent && eventname[(int)data[0]]) {
			const xEvent *e = (const xEvent *)data;

			if (dend < data + sz_xEvent)
				return;
			proto_print(context, data + 4,
			    "<%s #%u detail %u>",
			    eventname[e->u.u.type],
			    U16(e->u.u.sequenceNumber),
			    e->u.u.detail);

			switch (e->u.u.type) {
			case KeyPress:
			    proto_print(context, data + sz_xEvent,
				"  state %04x", 
				U16(e->u.keyButtonPointer.state));
			    break;
			default:
			    proto_print(context, data + sz_xEvent, "");
			}
			return;
		}

		if (dend < data + 32)
			return;
		proto_print(context, data + 32, "*** unknown server message");
		return;
	}

	/* Other: */
	proto_print(context, dend, "???");
}

struct proto proto_x11 = {
	6000, 				/* port */
	sizeof (struct x11state),	/* state */
	init,				/* init */
	decode,				/* decode */
	0				/* final */
};
